@@ -5,9 +5,6 @@
 
 . $(dirname $BASH_SOURCE)/functions.sh
 
-GLOBAL_OPTS="-M 128m"
-A_OPTS="--relaxed"
-
 if [ -n "$MSYSTEM" ]; then
   QUARTUS_ROOTDIR=$(cygpath -u $QUARTUS_ROOTDIR)
 fi
@@ -22,7 +19,12 @@ directory.  For example:
 
 EOF
   exit 1
-elif [ -d "$QUARTUS_ROOTDIR/eda/sim_lib" ]; then
+fi
+
+GLOBAL_OPTS="-M 128m"
+A_OPTS="--relaxed --relative=$QUARTUS_ROOTDIR"
+
+if [ -d "$QUARTUS_ROOTDIR/eda/sim_lib" ]; then
   src=$QUARTUS_ROOTDIR/eda/sim_lib
 
   echo "Using Quartus installation in $QUARTUS_ROOTDIR"
@@ -60,15 +62,19 @@ $src/sgate_pack.vhd
 $src/sgate.vhd
 EOF
 
+    if [ -f "$src/max_atoms.vhd" ]; then
     analyse_list max$(std_suffix $STD) <<EOF
 $src/max_atoms.vhd
 $src/max_components.vhd
 EOF
+    fi
 
+    if [ -f "$src/maxii_atoms.vhd" ]; then
     analyse_list maxii$(std_suffix $STD) <<EOF
 $src/maxii_atoms.vhd
 $src/maxii_components.vhd
 EOF
+    fi
 
     if [ -f "$src/maxv_atoms.vhd" ]; then
       analyse_list maxv$(std_suffix $STD) <<EOF
@@ -84,12 +90,28 @@ $src/fiftyfivenm_components.vhd
 EOF
     fi
 
+    if [ -f "$src/fourteennm_atoms.vhd" ]; then
+    analyse_list fourteennm$(std_suffix $STD) <<EOF
+$src/fourteennm_atoms.vhd
+$src/fourteennm_components.vhd
+EOF
+    fi
+
+    if [ -f "$src/tennm_atoms.vhd" ]; then
+    analyse_list tennm$(std_suffix $STD) <<EOF
+$src/tennm_atoms.vhd
+$src/tennm_components.vhd
+EOF
+    fi
+
+    if [ -f "$src/cycloneiv_atoms.vhd" ]; then
     analyse_list cycloneiv$(std_suffix $STD) <<EOF
 $src/cycloneiv_atoms.vhd
 $src/cycloneiv_components.vhd
 $src/cycloneiv_hssi_components.vhd
 $src/cycloneiv_hssi_atoms.vhd
 EOF
+    fi
 
     if [ -f "$src/cycloneive_atoms.vhd" ]; then
       analyse_list cycloneiv$(std_suffix $STD) <<EOF
@@ -120,12 +142,14 @@ $src/cyclone10lp_components.vhd
 EOF
     fi
 
+    if [ -f "$src/arriaii_atoms.vhd" ]; then
     analyse_list arriaii$(std_suffix $STD) <<EOF
 $src/arriaii_atoms.vhd
 $src/arriaii_components.vhd
 $src/arriaii_hssi_components.vhd
 $src/arriaii_hssi_atoms.vhd
 EOF
+    fi
 
     # produces compilation error
     #  analyse_list arriaii_pcie_hip$(std_suffix $STD) <<EOF
