@@ -33,9 +33,9 @@ echo
 src=$XILINX_VIVADO/data/vhdl/src
 
 GLOBAL_OPTS="-M 64m"
-A_OPTS="--relaxed"
+A_OPTS="--relaxed --relative=$XILINX_VIVADO"
 
-for STD in ${NVC_STD:-1993 2008}; do
+for STD in ${NVC_STD:-1993 2008 2019}; do
   analyse_list unisim$(std_suffix $STD) <<EOF
 $src/unisims/unisim_VPKG.vhd
 $src/unisims/unisim_retarget_VCOMP.vhd
