@@ -27,20 +27,6 @@
 #include <string.h>
 #include <time.h>
 
-static const char *get_relative_path(const char *path, const char *relative)
-{
-   if (relative != NULL && is_absolute_path(path)) {
-      const size_t rlen = strlen(relative);
-      if (strncmp(relative, path, rlen) == 0) {
-         path += rlen;
-         while (path[0] == DIR_SEP[0] || path[0] == '/')
-            path++;
-      }
-   }
-
-   return path;
-}
-
 ////////////////////////////////////////////////////////////////////////////////
 // Cobertura XML export
 
@@ -54,7 +40,7 @@ typedef struct {
 } cobertura_line_t;
 
 typedef struct _cobertura_class {
-   const char        *file;
+   char              *file;
    ident_t            name;
    cobertura_class_t *next;
    unsigned           nlines;
@@ -65,7 +51,7 @@ typedef struct _cobertura_class {
 typedef struct {
    hash_t            *class_map;
    cobertura_class_t *classes;
-   char              *relative;
+   const char        *relative;
 } cobertura_report_t;
 
 static cobertura_class_t *cobertura_get_class(cobertura_report_t *report,
@@ -73,9 +59,12 @@ static cobertura_class_t *cobertura_get_class(cobertura_report_t *report,
 {
    cobertura_class_t *c = hash_get(report->class_map, name);
    if (c == NULL) {
+      text_buf_t *tb = tb_new();
+      get_relative_path(tb, report->relative, loc_file_str(loc));
+
       c = xcalloc(sizeof(cobertura_class_t));
       c->name = name;
-      c->file = get_relative_path(loc_file_str(loc), report->relative);
+      c->file = tb_claim(tb);
       c->next = report->classes;
 
       report->classes = c;
@@ -116,28 +105,26 @@ static void cobertura_export_scope(cobertura_report_t *report,
                                    cobertura_class_t *class,
                                    cover_scope_t *s)
 {
-   if (cover_is_hier(s))
-      class = cobertura_get_class(report, s->block->block_name, &s->loc);
+   if (s->block_name != NULL)
+      class = cobertura_get_class(report, s->block_name, &s->loc);
 
    for (int i = 0; i < s->items.count; i++) {
-      cover_item_t *t = &(s->items.items[i]);
-      switch (t->kind) {
-      case COV_ITEM_STMT:
-         {
-            cobertura_line_t *l = cobertura_get_line(class, &(t->loc));
-            l->hits += t->data;
-         }
-         break;
-      case COV_ITEM_BRANCH:
-         {
-            cobertura_line_t *l = cobertura_get_line(class, &(t->loc));
+      const cover_item_t *item = s->items.items[i];
+      cobertura_line_t *l = cobertura_get_line(class, &(item->loc));
+
+      for (int j = 0; j < item->consecutive; j++) {
+         switch (item->kind) {
+         case COV_ITEM_STMT:
+            l->hits += item[j].data;
+            break;
+         case COV_ITEM_BRANCH:
             l->branch = true;
-            if (t->data > 0)
-               l->bflags |= t->flags;
+            if (item[j].data > 0)
+               l->bflags |= item[j].flags;
+            break;
+         default:
+            break;
          }
-         break;
-      default:
-         break;
       }
    }
 
@@ -208,7 +195,7 @@ void cover_export_cobertura(cover_data_t *data, FILE *f, const char *relative)
 {
    cobertura_report_t report = {
       .class_map = hash_new(64),
-      .relative = relative ? realpath(relative, NULL) : NULL,
+      .relative = relative,
    };
 
    cobertura_export_scope(&report, NULL, data->root_scope);
@@ -237,9 +224,9 @@ void cover_export_cobertura(cover_data_t *data, FILE *f, const char *relative)
            "line-rate=\"%f\" branch-rate=\"%f\" complexity=\"0.0\" "
            "lines-valid=\"%d\" lines-covered=\"%d\" "
            "branches-valid=\"%d\" branches-covered=\"%d\" "
-           "timestamp=\"%lu\">\n",
+           "timestamp=\"%llu\">\n",
            version, line_rate, branch_rate, nlines, hitlines, nbranches,
-           hitbranches, timestamp);
+           hitbranches, (long long)timestamp);
    fprintf(f, "<sources>\n");
    fprintf(f, "<source>.</source>\n");
    fprintf(f, "</sources>\n");
@@ -260,27 +247,60 @@ void cover_export_cobertura(cover_data_t *data, FILE *f, const char *relative)
    for (cobertura_class_t *it = report.classes, *tmp; it; it = tmp) {
       tmp = it->next;
       free(it->lines);
+      free(it->file);
       free(it);
    }
 
-   free(report.relative);
    hash_free(report.class_map);
 }
 
 ////////////////////////////////////////////////////////////////////////////////
 // XML dump format for debugging and testing
 
-static void dump_scope_xml(cover_scope_t *s, int indent, const loc_t *loc,
+static void dump_item_xml(const cover_item_t *item, int indent, FILE *f)
+{
+   for (int i = 0; i < item->consecutive; i++) {
+      switch (item->kind) {
+      case COV_ITEM_STMT:
+         fprintf(f, "%*s<statement hier=\"%s\" data=\"%d\"/>\n", indent + 2, "",
+                 istr(item[i].hier), item[i].data);
+         break;
+      case COV_ITEM_BRANCH:
+         fprintf(f, "%*s<branch hier=\"%s\" data=\"%d\"/>\n", indent + 2, "",
+                 istr(item[i].hier), item[i].data);
+         break;
+      case COV_ITEM_EXPRESSION:
+         fprintf(f, "%*s<expression hier=\"%s\" data=\"%d\"/>\n", indent + 2,
+                 "", istr(item[i].hier), item[i].data);
+         break;
+      case COV_ITEM_TOGGLE:
+         fprintf(f, "%*s<toggle hier=\"%s\" data=\"%d\"/>\n", indent + 2, "",
+                 istr(item[i].hier), item[i].data);
+         break;
+      case COV_ITEM_FUNCTIONAL:
+         fprintf(f, "%*s<functional hier=\"%s\" data=\"%d\"/>\n", indent + 2,
+                 "", istr(item[i].hier), item[i].data);
+         break;
+      case COV_ITEM_STATE:
+         fprintf(f, "%*s<state hier=\"%s\" data=\"%d\"/>\n", indent + 2,
+                 "", istr(item[i].hier), item[i].data);
+         break;
+      }
+   }
+}
+
+static void dump_scope_xml(const cover_scope_t *s, int indent, const loc_t *loc,
                            const char *relative, FILE *f)
 {
    fprintf(f, "%*s<scope name=\"%s\"", indent, "", istr(s->name));
 
-   if (cover_is_hier(s))
-      fprintf(f, " block_name=\"%s\"", istr(s->block->block_name));
+   if (s->block_name != NULL)
+      fprintf(f, " block_name=\"%s\"", istr(s->block_name));
 
    if (s->loc.file_ref != FILE_INVALID && s->loc.file_ref != loc->file_ref) {
-      const char *path = get_relative_path(loc_file_str(&s->loc), relative);
-      fprintf(f, " file=\"%s\"", path);
+      LOCAL_TEXT_BUF tb = tb_new();
+      get_relative_path(tb, relative, loc_file_str(&s->loc));
+      fprintf(f, " file=\"%s\"", tb_get(tb));
    }
 
    if (s->loc.first_line != LINE_INVALID && s->loc.first_line > 0
@@ -289,35 +309,8 @@ static void dump_scope_xml(cover_scope_t *s, int indent, const loc_t *loc,
 
    fprintf(f, ">\n");
 
-   for (int i = 0; i < s->items.count; i++) {
-      cover_item_t *item = &(s->items.items[i]);
-      switch (item->kind) {
-      case COV_ITEM_STMT:
-         fprintf(f, "%*s<statement hier=\"%s\" data=\"%d\"/>\n", indent + 2, "",
-                 istr(item->hier), item->data);
-         break;
-      case COV_ITEM_BRANCH:
-         fprintf(f, "%*s<branch hier=\"%s\" data=\"%d\"/>\n", indent + 2, "",
-                 istr(item->hier), item->data);
-         break;
-      case COV_ITEM_EXPRESSION:
-         fprintf(f, "%*s<expression hier=\"%s\" data=\"%d\"/>\n", indent + 2,
-                 "", istr(item->hier), item->data);
-         break;
-      case COV_ITEM_TOGGLE:
-         fprintf(f, "%*s<toggle hier=\"%s\" data=\"%d\"/>\n", indent + 2, "",
-                 istr(item->hier), item->data);
-         break;
-      case COV_ITEM_FUNCTIONAL:
-         fprintf(f, "%*s<functional hier=\"%s\" data=\"%d\"/>\n", indent + 2,
-                 "", istr(item->hier), item->data);
-         break;
-      case COV_ITEM_STATE:
-         fprintf(f, "%*s<state hier=\"%s\" data=\"%d\"/>\n", indent + 2,
-                 "", istr(item->hier), item->data);
-         break;
-      }
-   }
+   for (int i = 0; i < s->items.count; i++)
+      dump_item_xml(s->items.items[i], indent, f);
 
    for (int i = 0; i < s->children.count; i++)
       dump_scope_xml(s->children.items[i], indent + 2, &s->loc, relative, f);
