@@ -20,8 +20,9 @@
 #include "common.h"
 #include "diag.h"
 #include "jit/jit-exits.h"
-#include "jit/jit-priv.h"
 #include "jit/jit-ffi.h"
+#include "jit/jit-priv.h"
+#include "printf.h"
 #include "rt/mspace.h"
 #include "tree.h"
 #include "type.h"
@@ -88,7 +89,7 @@ static void interp_dump_reg(jit_interp_t *state, int64_t ival)
       void *base;
       if ((base = mspace_find(state->mspace, (void *)(intptr_t)ival, &size))) {
          if (size == 0)
-            color_printf("$!red$ ==> bad mspace object$$\n");
+            nvc_printf("$!red$ ==> bad mspace object$$\n");
          else {
             printf(" ==> %zu byte mspace object\n", size);
             jit_hexdump(base, size, 8, (void *)(intptr_t)ival, "\t\t\t");
@@ -177,8 +178,6 @@ static inline void *interp_get_pointer(jit_interp_t *state, jit_value_t value)
       return state->regs[value.reg].pointer + value.disp;
    case JIT_ADDR_ABS:
       return (void *)(intptr_t)value.int64;
-   case JIT_ADDR_COVER:
-      return jit_get_cover_ptr(state->func, value);
    default:
       CANNOT_HANDLE(value);
    }
@@ -212,10 +211,6 @@ static inline jit_scalar_t interp_get_scalar(jit_interp_t *state,
       };
    case JIT_ADDR_ABS:
       return (jit_scalar_t){ .pointer = (void *)(intptr_t)value.int64 };
-   case JIT_ADDR_COVER:
-      return (jit_scalar_t){
-         .pointer = jit_get_cover_ptr(state->func, value)
-      };
    default:
       CANNOT_HANDLE(value);
    }
@@ -886,6 +881,9 @@ static void interp_sadd(jit_interp_t *state, jit_ir_t *ir)
    const void *ptr = interp_get_pointer(state, ir->arg1);
    const int64_t addend = interp_get_int(state, ir->arg2);
 
+   JIT_ASSERT(ptr != NULL);
+   JIT_ASSERT((intptr_t)ptr >= 4096);
+
 #define SADD(type) do {                                         \
       u##type cur = *(u##type *)ptr;                            \
       *(u##type *)ptr = saturate_add(cur, addend);              \
