@@ -286,6 +286,7 @@ typedef unsigned valnum_t;
 typedef struct {
    jit_ir_t *ir;
    valnum_t  vn;
+   int       tuple[3];
 } lvn_tab_t;
 
 typedef struct {
@@ -463,38 +464,36 @@ static void jit_lvn_generic(jit_ir_t *ir, lvn_state_t *state, valnum_t vn)
          tab->ir = ir;
          tab->vn = state->regvn[ir->result] =
             (vn == VN_INVALID ? lvn_new_value(state) : vn);
-         break;
+         memcpy(tab->tuple, tuple, sizeof(tuple));
+         return;
       }
       else if (tab->vn != state->regvn[tab->ir->result]) {
          // Stale entry may be reused if we do not find matching value
          stale = idx;
          continue;
       }
-      else {
-         int cmp[3];
-         lvn_get_tuple(tab->ir, state, cmp);
+      else if (memcmp(tuple, tab->tuple, sizeof(tuple)) == 0) {
+         assert(tab->ir->result != JIT_REG_INVALID);
 
-         if (cmp[0] == tuple[0] && cmp[1] == tuple[1] && cmp[2] == tuple[2]) {
-            assert(tab->ir->result != JIT_REG_INVALID);
+         ir->op   = J_MOV;
+         ir->size = JIT_SZ_UNSPEC;
+         ir->cc   = JIT_CC_NONE;
 
-            ir->op   = J_MOV;
-            ir->size = JIT_SZ_UNSPEC;
-            ir->cc   = JIT_CC_NONE;
+         // Propagate constants where possible
+         int64_t cval;
+         if (lvn_get_const(state->regvn[tab->ir->result], state, &cval))
+            ir->arg1 = LVN_CONST(cval);
+         else
+            ir->arg1 = LVN_REG(tab->ir->result);
 
-            // Propagate constants where possible
-            int64_t cval;
-            if (lvn_get_const(state->regvn[tab->ir->result], state, &cval))
-               ir->arg1 = LVN_CONST(cval);
-            else
-               ir->arg1 = LVN_REG(tab->ir->result);
+         ir->arg2.kind = JIT_VALUE_INVALID;
 
-            ir->arg2.kind = JIT_VALUE_INVALID;
-
-            state->regvn[ir->result] = tab->vn;
-            break;
-         }
+         state->regvn[ir->result] = tab->vn;
+         return;
       }
    }
+
+   state->regvn[ir->result] = VN_INVALID;   // Reached iteration limit
 }
 
 static void jit_lvn_mul(jit_ir_t *ir, lvn_state_t *state)
@@ -665,7 +664,7 @@ static void jit_lvn_xor(jit_ir_t *ir, lvn_state_t *state)
 static void jit_lvn_shl(jit_ir_t *ir, lvn_state_t *state)
 {
    int64_t lhs, rhs;
-   if (lvn_can_fold(ir, state, &lhs, &rhs) && rhs >= 0 && rhs < 64)
+   if (lvn_can_fold(ir, state, &lhs, &rhs) && rhs >= 0 && rhs < 64 && lhs >= 0)
       lvn_convert_mov(ir, state, LVN_CONST(lhs << rhs));
    else if (lvn_is_const(ir->arg2, state, &rhs) && rhs == 0)
       lvn_convert_mov(ir, state, ir->arg1);
@@ -752,6 +751,8 @@ static void jit_lvn_csel(jit_ir_t *ir, lvn_state_t *state)
    const int fconst = state->regvn[state->func->nregs];
    if (fconst != VN_INVALID)
       lvn_convert_mov(ir, state, fconst ? ir->arg1 : ir->arg2);
+   else
+      state->regvn[ir->result] = VN_INVALID;
 }
 
 static void jit_lvn_cset(jit_ir_t *ir, lvn_state_t *state)
@@ -759,6 +760,8 @@ static void jit_lvn_cset(jit_ir_t *ir, lvn_state_t *state)
    const int fconst = state->regvn[state->func->nregs];
    if (fconst != VN_INVALID)
       lvn_convert_mov(ir, state, LVN_CONST(fconst));
+   else
+      state->regvn[ir->result] = VN_INVALID;
 }
 
 static void jit_lvn_jump(jit_ir_t *ir, lvn_state_t *state)
