@@ -1,5 +1,5 @@
 //
-//  Copyright (C) 2024-2025  Nick Gasson
+//  Copyright (C) 2024-2026  Nick Gasson
 //
 //  This program is free software: you can redistribute it and/or modify
 //  it under the terms of the GNU General Public License as published by
@@ -107,6 +107,15 @@ STATIC_ASSERT(sizeof(mir_stamp_t) == 4);
          (mir_type_t){ .bits = (p).bits };      \
       })
 
+typedef struct {
+   uint16_t first;
+   uint16_t width;
+} mir_vreg_t;
+
+STATIC_ASSERT(sizeof(mir_vreg_t) == 4);
+
+#define MIR_VREG_MAX UINT16_MAX
+
 typedef enum {
    MIR_CMP_EQ,
    MIR_CMP_NEQ,
@@ -259,6 +268,8 @@ typedef enum {
    MIR_OP_SCHED_INACTIVE,
    MIR_OP_SCHED_DEPOSIT,
    MIR_OP_PUT_DRIVER,
+   MIR_OP_TABLE_REF,
+   MIR_OP_GET_COUNTERS,
 } mir_op_t;
 
 typedef enum {
@@ -311,6 +322,7 @@ typedef enum {
 
 typedef enum {
    MIR_MEM_NONE,
+   MIR_MEM_NULL,
    MIR_MEM_CONST,
    MIR_MEM_STACK,
    MIR_MEM_LOCAL,
@@ -342,8 +354,8 @@ typedef enum {
    MIR_VEC_MOD,
    MIR_VEC_SLL,
    MIR_VEC_SRL,
-   MIR_VEC_SLA,
    MIR_VEC_SRA,
+   MIR_VEC_EXP,
 } mir_vec_op_t;
 
 #define MIR_APPEND UINT_MAX
@@ -376,11 +388,13 @@ mir_type_t mir_file_type(mir_unit_t *mu, mir_type_t base);
 mir_type_t mir_vec2_type(mir_unit_t *mu, int size, bool issigned);
 mir_type_t mir_vec4_type(mir_unit_t *mu, int size, bool issigned);
 
+mir_type_t mir_vector_slice(mir_unit_t *mu, mir_type_t base, int size);
 mir_type_t mir_get_base(mir_unit_t *mu, mir_type_t type);
 mir_type_t mir_get_elem(mir_unit_t *mu, mir_type_t type);
 mir_type_t mir_get_pointer(mir_unit_t *mu, mir_type_t type);
 unsigned mir_get_dims(mir_unit_t *mu, mir_type_t type);
 unsigned mir_get_size(mir_unit_t *mu, mir_type_t type);
+unsigned mir_get_slots(mir_unit_t *mu, mir_type_t type);
 bool mir_get_signed(mir_unit_t *mu, mir_type_t type);
 mir_class_t mir_get_class(mir_unit_t *mu, mir_type_t type);
 mir_repr_t mir_get_repr(mir_unit_t *mu, mir_type_t type);
@@ -403,6 +417,7 @@ unsigned mir_count_nodes(mir_unit_t *mu, mir_block_t block);
 mir_value_t mir_get_node(mir_unit_t *mu, mir_block_t block, unsigned nth);
 unsigned mir_count_vars(mir_unit_t *mu);
 unsigned mir_count_params(mir_unit_t *mu);
+unsigned mir_count_vregs(mir_unit_t *mu);
 
 mir_op_t mir_get_op(mir_unit_t *mu, mir_value_t node);
 unsigned mir_count_args(mir_unit_t *mu, mir_value_t node);
@@ -429,6 +444,7 @@ mir_type_t mir_get_var_type(mir_unit_t *mu, mir_value_t value);
 void mir_set_input(mir_unit_t *mu, mir_value_t phi, unsigned nth,
                    mir_block_t block, mir_value_t value);
 bool mir_block_finished(mir_unit_t *mu, mir_block_t block);
+mir_vreg_t mir_get_vreg(mir_unit_t *mu, mir_value_t value);
 
 void mir_set_result(mir_unit_t *mu, mir_type_t type);
 mir_type_t mir_get_result(mir_unit_t *mu);
@@ -553,6 +569,9 @@ mir_value_t mir_build_all(mir_unit_t *mu, mir_value_t access);
 mir_value_t mir_build_address_of(mir_unit_t *mu, mir_value_t array);
 mir_value_t mir_build_array_ref(mir_unit_t *mu, mir_value_t array,
                                 mir_value_t offset);
+mir_value_t mir_build_table_ref(mir_unit_t *mu, mir_value_t array,
+                                mir_value_t stride, const mir_value_t *args,
+                                int nargs);
 mir_value_t mir_build_record_ref(mir_unit_t *mu, mir_value_t record,
                                  unsigned field);
 
@@ -677,9 +696,10 @@ void mir_build_transfer_signal(mir_unit_t *mu, mir_value_t target,
                                mir_value_t reject, mir_value_t after);
 
 // Coverage
-void mir_build_cover_stmt(mir_unit_t *mu, uint32_t tag);
-void mir_build_cover_branch(mir_unit_t *mu, uint32_t tag);
-void mir_build_cover_expr(mir_unit_t *mu, uint32_t tag);
+mir_value_t mir_build_get_counters(mir_unit_t *mu, ident_t block);
+void mir_build_cover_stmt(mir_unit_t *mu, mir_value_t counters, uint32_t tag);
+void mir_build_cover_branch(mir_unit_t *mu, mir_value_t counters, uint32_t tag);
+void mir_build_cover_expr(mir_unit_t *mu, mir_value_t counters, uint32_t tag);
 void mir_build_cover_toggle(mir_unit_t *mu, mir_value_t signal, uint32_t tag);
 void mir_build_cover_state(mir_unit_t *mu, mir_value_t signal, mir_value_t low,
                            uint32_t tag);
@@ -717,14 +737,14 @@ void mir_build_add_trigger(mir_unit_t *mu, mir_value_t trigger);
 
 // Linking
 mir_value_t mir_build_link_package(mir_unit_t *mu, ident_t name);
-mir_value_t mir_build_link_var(mir_unit_t *mu, ident_t unit,
-                               mir_value_t context, ident_t name,
-                               mir_type_t type);
+mir_value_t mir_build_link_var(mir_unit_t *mu, mir_value_t context,
+                               ident_t name, mir_type_t type);
 void mir_build_bind_foreign(mir_unit_t *mu, mir_value_t spec,
                             mir_value_t length, mir_value_t locus);
 mir_value_t mir_build_bind_external(mir_unit_t *mu, mir_value_t locus,
                                     ident_t scope, mir_type_t type,
-                                    mir_stamp_t stamp);
+                                    mir_stamp_t stamp, const mir_value_t *args,
+                                    int nargs);
 mir_value_t mir_build_context_upref(mir_unit_t *mu, int hops);
 mir_value_t mir_build_var_upref(mir_unit_t *mu, int hops, int nth);
 
