@@ -1,5 +1,5 @@
 //
-//  Copyright (C) 2024-2025  Nick Gasson
+//  Copyright (C) 2024-2026  Nick Gasson
 //
 //  This program is free software: you can redistribute it and/or modify
 //  it under the terms of the GNU General Public License as published by
@@ -16,12 +16,14 @@
 //
 
 #include "util.h"
+#include "array.h"
 #include "ident.h"
 #include "mask.h"
 #include "mir/mir-node.h"
 #include "mir/mir-priv.h"
 #include "mir/mir-structs.h"
 #include "option.h"
+#include "printf.h"
 
 #include <assert.h>
 #include <stdlib.h>
@@ -46,11 +48,9 @@ typedef struct {
    cfg_edge_list_t in;
    cfg_edge_list_t out;
    bit_mask_t      dom;
-#if 0
    bit_mask_t      livein;
    bit_mask_t      varkill;
    bit_mask_t      liveout;
-#endif
 } cfg_block_t;
 
 typedef uint32_t valnum_t;
@@ -76,14 +76,20 @@ typedef struct {
 } gvn_state_t;
 
 typedef struct {
-   bit_mask_t visited;
-   uint8_t    uses[];
-} dce_state_t;
+   mir_value_t value;
+   unsigned    first;
+   unsigned    last;
+} live_range_t;
+
+typedef struct {
+   unsigned     *block_start;
+   live_range_t  lrg[];
+} ra_state_t;
 
 typedef struct {
    cfg_block_t *cfg;
    gvn_state_t *gvn;
-   dce_state_t *dce;
+   ra_state_t  *ra;
 } mir_optim_t;
 
 static void mir_dump_optim(mir_unit_t *mu, mir_optim_t *an);
@@ -128,40 +134,68 @@ static mir_block_t cfg_get_edge(const cfg_edge_list_t *list, int nth)
       return list->u.external[nth];
 }
 
-static cfg_block_t *mir_get_cfg(mir_unit_t *mu)
+static void cfg_walk_block(mir_unit_t *mu, mir_block_t block, cfg_block_t *cfg,
+                           bit_mask_t *visited)
 {
-   cfg_block_t *cfg = xcalloc_array(mu->blocks.count, sizeof(cfg_block_t));
-   cfg[0].entry = true;
+   if (mask_test_and_set(visited, block.id))
+      return;
 
-   if (mir_get_kind(mu) == MIR_UNIT_PROCESS && mu->blocks.count > 1)
-      cfg[1].entry = true;
+   DEBUG_ONLY(mir_set_cursor(mu, block, MIR_APPEND));
 
-   for (int i = 0; i < mu->blocks.count; i++) {
-      const block_data_t *bd = &(mu->blocks.items[i]);
-      mir_block_t this = { .tag = MIR_TAG_BLOCK, .id = i };
+   cfg[block.id].block = block;
 
-      DEBUG_ONLY(mir_set_cursor(mu, this, MIR_APPEND));
-      MIR_ASSERT(bd->num_nodes > 0, "empty basic block");
+   const block_data_t *bd = &(mu->blocks.items[block.id]);
+   MIR_ASSERT(bd->num_nodes > 0, "empty basic block");
 
-      const node_data_t *last = &(mu->nodes[bd->nodes[bd->num_nodes - 1]]);
-      MIR_ASSERT(mir_is_terminator(last->op),
-                 "last operation in block is not terminator");
+   const node_data_t *last = &(mu->nodes[bd->nodes[bd->num_nodes - 1]]);
+   MIR_ASSERT(mir_is_terminator(last->op),
+              "last operation in block is not terminator");
 
-      switch (last->op) {
-      case MIR_OP_WAIT:
-         assert(last->args[0].tag == MIR_TAG_BLOCK);
-         cfg[last->args[0].id].entry = true;
-         // Fall-through
-      case MIR_OP_RETURN:
-         cfg[i].returns = true;
-         break;
-      default:
-         assert(last->nargs <= MIR_INLINE_ARGS);
+   switch (last->op) {
+   case MIR_OP_WAIT:
+      assert(last->args[0].tag == MIR_TAG_BLOCK);
+      cfg[last->args[0].id].entry = true;
+      cfg_walk_block(mu, mir_cast_block(last->args[0]), cfg, visited);
+      // Fall-through
+   case MIR_OP_RETURN:
+      cfg[block.id].returns = true;
+      break;
+   default:
+      {
+         const mir_value_t *args = mir_get_args(mu, last);
          for (int j = 0; j < last->nargs; j++) {
-            if (last->args[j].tag == MIR_TAG_BLOCK)
-               cfg_add_edge(cfg, this, mir_cast_block(last->args[j]));
+            if (args[j].tag == MIR_TAG_BLOCK) {
+               mir_block_t target = mir_cast_block(args[j]);
+               cfg_add_edge(cfg, block, target);
+               cfg_walk_block(mu, target, cfg, visited);
+            }
          }
       }
+      break;
+   }
+}
+
+static cfg_block_t *mir_get_cfg(mir_unit_t *mu)
+{
+   LOCAL_BIT_MASK visited;
+   mask_init(&visited, mu->blocks.count);
+
+   cfg_block_t *cfg = xcalloc_array(mu->blocks.count, sizeof(cfg_block_t));
+   for (int i = 0; i < mu->blocks.count; i++) {
+      mir_block_t this = { .tag = MIR_TAG_BLOCK, .id = i };
+      cfg[i].block = this;
+   }
+
+
+   cfg[0].entry = true;
+   cfg_walk_block(mu, cfg[0].block, cfg, &visited);
+
+   const bool multi_entry =
+      mu->kind == MIR_UNIT_PROCESS || mu->kind == MIR_UNIT_PROPERTY;
+
+   if (multi_entry && mu->blocks.count > 1) {
+      cfg[1].entry = true;
+      cfg_walk_block(mu, cfg[1].block, cfg, &visited);
    }
 
    return cfg;
@@ -175,6 +209,9 @@ static void mir_free_cfg(mir_unit_t *mu, cfg_block_t *cfg)
       if (cb->out.max > CFG_EMBED_EDGES) free(cb->out.u.external);
 
       mask_free(&cb->dom);
+      mask_free(&cb->livein);
+      mask_free(&cb->varkill);
+      mask_free(&cb->liveout);
    }
 
    free(cfg);
@@ -226,12 +263,11 @@ static void mir_dominator_tree(mir_unit_t *mu, cfg_block_t *cfg)
 ////////////////////////////////////////////////////////////////////////////////
 // Liveness analysis
 
-#if 0
 static void mir_do_liveness(mir_unit_t *mu, mir_optim_t *opt)
 {
    // Algorithm from "Engineering a Compiler" chapter 8.6
 
-   for (int i = 0; i < mu->num_blocks; i++) {
+   for (int i = 0; i < mu->blocks.count; i++) {
       cfg_block_t *cb = &(opt->cfg[i]);
       mask_init(&cb->livein, mu->num_nodes);
       mask_init(&cb->varkill, mu->num_nodes);
@@ -239,13 +275,13 @@ static void mir_do_liveness(mir_unit_t *mu, mir_optim_t *opt)
 
       const block_data_t *bd = mir_block_data(mu, cb->block);
       for (int j = 0; j < bd->num_nodes; j++) {
-         mir_value_t node = { .tag = MIR_TAG_NODE, .id = bd->nodes[i] };
+         mir_value_t node = { .tag = MIR_TAG_NODE, .id = bd->nodes[j] };
          const node_data_t *n = mir_node_data(mu, node);
 
+         mask_set(&cb->varkill, node.id);
+
          if (n->op == MIR_OP_CONST || n->op == MIR_OP_CONST_REAL)
-            continue;
-         else if (mir_is_null(n->type))
-            continue;   // Produces no value
+            continue;   // Arguments have special encoding
 
          const mir_value_t *args = mir_get_args(mu, n);
          for (int k = 0; k < n->nargs; k++) {
@@ -253,8 +289,6 @@ static void mir_do_liveness(mir_unit_t *mu, mir_optim_t *opt)
             if (arg.tag == MIR_TAG_NODE && !mask_test(&cb->varkill, arg.id))
                mask_set(&cb->livein, arg.id);
          }
-
-         mask_set(&cb->varkill, node.id);
       }
    }
 
@@ -266,7 +300,7 @@ static void mir_do_liveness(mir_unit_t *mu, mir_optim_t *opt)
    do {
       changed = false;
 
-      for (int i = mu->num_blocks - 1; i >= 0; i--) {
+      for (int i = mu->blocks.count - 1; i >= 0; i--) {
          cfg_block_t *cb = &(opt->cfg[i]);
          mask_clearall(&new);
 
@@ -286,7 +320,7 @@ static void mir_do_liveness(mir_unit_t *mu, mir_optim_t *opt)
    } while (changed);
 
    // Replaced "upward exposed variables" set with live-in
-   for (int i = 0; i < mu->num_blocks; i++) {
+   for (int i = 0; i < mu->blocks.count; i++) {
       cfg_block_t *cb = &(opt->cfg[i]);
       mask_copy(&tmp, &cb->liveout);
       mask_subtract(&tmp, &cb->varkill);
@@ -296,7 +330,6 @@ static void mir_do_liveness(mir_unit_t *mu, mir_optim_t *opt)
    mask_free(&new);
    mask_free(&tmp);
 }
-#endif
 
 ////////////////////////////////////////////////////////////////////////////////
 // Global value numbering
@@ -601,6 +634,15 @@ static void gvn_visit_block(mir_unit_t *mu, mir_block_t block,
       case MIR_OP_CONTEXT_UPREF:
       case MIR_OP_UNWRAP:
       case MIR_OP_INSERT:
+      case MIR_OP_NULL:
+      case MIR_OP_FUNCTION_TRIGGER:
+      case MIR_OP_LEVEL_TRIGGER:
+      case MIR_OP_OR_TRIGGER:
+      case MIR_OP_ARRAY_REF:
+      case MIR_OP_TABLE_REF:
+      case MIR_OP_TEST:
+      case MIR_OP_EXTRACT:
+      case MIR_OP_NEG:
          gvn_generic(mu, node, block, opt);
          break;
       case MIR_OP_AND:
@@ -618,6 +660,9 @@ static void gvn_visit_block(mir_unit_t *mu, mir_block_t block,
          break;
       case MIR_OP_INIT_SIGNAL:
       case MIR_OP_PORT_CONVERSION:
+      case MIR_OP_PROTECTED_INIT:
+      case MIR_OP_FCALL:
+      case MIR_OP_SYSCALL:
          opt->gvn->nodevn[node.id] = gvn_new_value(node, opt->gvn);
          break;
       default:
@@ -681,70 +726,25 @@ static void mir_do_gvn(mir_unit_t *mu, mir_optim_t *opt)
 }
 
 ////////////////////////////////////////////////////////////////////////////////
-// Simple dead code elimination by counting uses
-
-static void dce_visit_block(mir_unit_t *mu, mir_block_t block, mir_optim_t *opt)
-{
-   if (mask_test(&opt->dce->visited, block.id))
-      return;
-
-   mask_set(&opt->dce->visited, block.id);
-
-   const block_data_t *bd = mir_block_data(mu, block);
-   for (int i = 0; i < bd->num_nodes; i++) {
-      mir_value_t node = { .tag = MIR_TAG_NODE, .id = bd->nodes[i] };
-      node_data_t *n = mir_node_data(mu, node);
-      const mir_value_t *args = mir_get_args(mu, n);
-
-      for (int j = 0; j < n->nargs; j++) {
-         mir_value_t arg = args[j];
-         if (arg.tag == MIR_TAG_NODE)
-            opt->dce->uses[arg.id] = saturate_add(opt->dce->uses[arg.id], 1);
-      }
-
-      if (mir_is_null(n->type) || n->op == MIR_OP_STORE)
-         opt->dce->uses[node.id] = 1;   // Does not produce value
-   }
-
-   cfg_block_t *cb = &(opt->cfg[block.id]);
-   for (int i = 0; i < cb->out.count; i++)
-      dce_visit_block(mu, cfg_get_edge(&cb->out, i), opt);
-}
+// Dead code elimination using liveness information
 
 static void mir_do_dce(mir_unit_t *mu, mir_optim_t *opt)
 {
-   opt->dce = xcalloc_flex(sizeof(dce_state_t), mu->num_nodes, sizeof(uint8_t));
-   mask_init(&opt->dce->visited, mu->num_nodes);
-
-   for (int i = 0; i < mu->blocks.count; i++) {
-      if (opt->cfg[i].entry) {
-         mir_block_t this = { .tag = MIR_TAG_BLOCK, .id = i };
-         dce_visit_block(mu, this, opt);
-      }
-   }
-
-   if (opt_get_verbose(OPT_DCE_VERBOSE, istr(mu->name)))
-      mir_dump_optim(mu, opt);
+   bit_mask_t live;
+   mask_init(&live, mu->num_nodes);
 
    for (int i = mu->blocks.count - 1; i >= 0; i--) {
       mir_block_t this = { .tag = MIR_TAG_BLOCK, .id = i };
       const block_data_t *bd = mir_block_data(mu, this);
 
+      mask_copy(&live, &(opt->cfg[i].liveout));
+
       for (int j = bd->num_nodes - 1; j >= 0; j--) {
          mir_value_t node = { .tag = MIR_TAG_NODE, .id = bd->nodes[j] };
-         if (opt->dce->uses[node.id] == 0) {
-            const node_data_t *nd = mir_node_data(mu, node);
-            const mir_value_t *args = mir_get_args(mu, nd);
-            for (int k = 0; k < nd->nargs; k++) {
-               if (args[k].tag == MIR_TAG_NODE) {
-                  const uint8_t uses = opt->dce->uses[args[k].id];
-                  assert(uses > 0);
-                  if (uses < UINT8_MAX)
-                     opt->dce->uses[args[k].id] = uses - 1;
-               }
-            }
+         node_data_t *n = mir_node_data(mu, node);
 
-            DEBUG_ONLY(const mir_op_t op = nd->op);
+         if (!mir_is_null(n->type) && !mask_test(&live, node.id)) {
+            DEBUG_ONLY(const mir_op_t op = n->op);
 
             mir_set_cursor(mu, this, j);
             mir_delete(mu);
@@ -752,14 +752,271 @@ static void mir_do_dce(mir_unit_t *mu, mir_optim_t *opt)
             DEBUG_ONLY(mir_comment(mu, "Dead %s definition of %%%u",
                                    mir_op_string(op), node.id));
          }
+
+         mask_clear(&live, node.id);
+
+         const mir_value_t *args = mir_get_args(mu, n);
+
+         for (int k = 0; k < n->nargs; k++) {
+            mir_value_t arg = args[k];
+            if (arg.tag == MIR_TAG_NODE)
+               mask_set(&live, arg.id);
+         }
       }
    }
 
+   if (opt_get_verbose(OPT_DCE_VERBOSE, istr(mu->name)))
+      mir_dump_optim(mu, opt);
+
    mir_compact(mu);
 
-   mask_free(&opt->dce->visited);
-   free(opt->dce);
-   opt->dce = NULL;
+   mask_free(&live);
+}
+
+////////////////////////////////////////////////////////////////////////////////
+// Control flow graph cleanup
+
+static void mir_do_cfg_cleanup(mir_unit_t *mu, mir_optim_t *opt)
+{
+   for (int i = 0; i < mu->blocks.count; i++) {
+      if (!opt->cfg[i].entry && opt->cfg[i].in.count == 0) {
+         block_data_t *bd = &(mu->blocks.items[i]);
+
+         // TODO: could delete the block instead
+         mir_set_cursor(mu, opt->cfg[i].block, 0);
+         mir_delete(mu);
+         mir_build_unreachable(mu, MIR_NULL_VALUE);
+
+         bd->num_nodes = 1;
+      }
+   }
+
+   if (opt_get_verbose(OPT_CFG_VERBOSE, istr(mu->name)))
+      mir_dump_optim(mu, opt);
+}
+
+////////////////////////////////////////////////////////////////////////////////
+// Virtual register allocation
+
+static void ra_grow_range(mir_value_t value, unsigned pos, ra_state_t *ra)
+{
+   assert(value.tag == MIR_TAG_NODE);
+
+   live_range_t *lrg = &(ra->lrg[value.id]);
+   assert(mir_is_null(lrg->value) || mir_equals(lrg->value, value));
+
+   lrg->value = value;
+   lrg->first = MIN(lrg->first, pos);
+   lrg->last = MAX(lrg->last, pos);
+}
+
+static void ra_build_lrg(mir_unit_t *mu, mir_optim_t *opt, mir_block_t block,
+                         bit_mask_t *visited)
+{
+   if (mask_test_and_set(visited, block.id))
+      return;
+
+   const cfg_block_t *cb = &(opt->cfg[block.id]);
+
+   const unsigned first = opt->ra->block_start[block.id];
+   assert(first != UINT_MAX);
+
+   for (size_t bit = -1; mask_iter(&cb->livein, &bit);) {
+      mir_value_t node = { .tag = MIR_TAG_NODE, .id = bit };
+      ra_grow_range(node, first, opt->ra);
+   }
+
+   const block_data_t *bd = mir_block_data(mu, block);
+
+   for (int i = 0; i < bd->num_nodes; i++) {
+      mir_value_t node = { .tag = MIR_TAG_NODE, .id = bd->nodes[i] };
+      node_data_t *n = mir_node_data(mu, node);
+
+      if (!mir_is_null(n->type))
+         ra_grow_range(node, first + i, opt->ra);
+
+      const mir_value_t *args = mir_get_args(mu, n);
+
+      for (int j = 0; j < n->nargs; j++) {
+         mir_value_t arg = args[j];
+         if (arg.tag == MIR_TAG_NODE)
+            ra_grow_range(arg, first + i, opt->ra);
+      }
+   }
+
+   for (size_t bit = -1; mask_iter(&cb->liveout, &bit); ) {
+      mir_value_t node = { .tag = MIR_TAG_NODE, .id = bit };
+      ra_grow_range(node, first + bd->num_nodes - 1, opt->ra);
+   }
+
+   for (int i = 0; i < cb->out.count; i++) {
+      mir_block_t next = cfg_get_edge(&(cb->out), i);
+      ra_build_lrg(mu, opt, next, visited);
+   }
+}
+
+static void ra_compute_postorder(mir_optim_t *opt, mir_block_t block,
+                                 bit_mask_t *visited, mir_block_t **tailp)
+{
+   if (mask_test_and_set(visited, block.id))
+      return;
+
+   const cfg_block_t *cb = &(opt->cfg[block.id]);
+
+   for (int i = 0; i < cb->out.count; i++) {
+      mir_block_t next = cfg_get_edge(&(cb->out), i);
+      ra_compute_postorder(opt, next, visited, tailp);
+   }
+
+   *(*tailp)++ = block;
+}
+
+static int live_range_cmp(const void *a, const void *b)
+{
+   const live_range_t *la = a;
+   const live_range_t *lb = b;
+
+   if (la->first < lb->first)
+      return -1;
+   else if (la->first > lb->first)
+      return 1;
+   else
+      return 0;
+}
+
+static inline void ra_shift_active(live_range_t **active, unsigned to,
+                                   unsigned from, unsigned count)
+{
+   if (to != from && count == 1)
+      active[to] = active[from];
+   else if (to != from && count > 1)
+      memmove(active + to, active + from, count * sizeof(live_range_t *));
+}
+
+static void mir_do_ra(mir_unit_t *mu, mir_optim_t *opt)
+{
+   opt->ra = xmalloc_flex(sizeof(ra_state_t), mu->num_nodes,
+                          sizeof(live_range_t));
+
+   const live_range_t null_lrg = { MIR_NULL_VALUE, UINT_MAX, 0 };
+   for (int i = 0; i < mu->num_nodes; i++)
+      opt->ra->lrg[i] = null_lrg;
+
+   bit_mask_t visited;
+   mask_init(&visited, mu->blocks.count);
+
+   mir_block_t *postorder LOCAL =
+      xmalloc_array(mu->blocks.count, sizeof(mir_block_t));
+   mir_block_t *tailp = postorder;
+
+   for (int i = 0; i < mu->blocks.count; i++) {
+      if (opt->cfg[i].entry) {
+         mir_block_t this = { .tag = MIR_TAG_BLOCK, .id = i };
+         assert(!mask_test(&visited, i));
+         ra_compute_postorder(opt, this, &visited, &tailp);
+      }
+   }
+   assert(tailp <= postorder + mu->blocks.count);
+
+   opt->ra->block_start = xmalloc_array(mu->blocks.count, sizeof(unsigned));
+   for (int i = 0; i < mu->blocks.count; i++)
+      opt->ra->block_start[i] = UINT_MAX;
+
+   // Number nodes in reverse post-order
+   for (int i = tailp - postorder - 1, index = 0; i >= 0; i--) {
+      mir_block_t this = postorder[i];
+      opt->ra->block_start[this.id] = index;
+      index += mir_block_data(mu, this)->num_nodes;
+   }
+
+   mask_clearall(&visited);
+
+   for (int i = 0; i < mu->blocks.count; i++) {
+      if (opt->cfg[i].entry) {
+         mir_block_t this = { .tag = MIR_TAG_BLOCK, .id = i };
+         assert(!mask_test(&visited, i));
+         ra_build_lrg(mu, opt, this, &visited);
+      }
+   }
+
+   mask_free(&visited);
+
+   qsort(opt->ra->lrg, mu->num_nodes, sizeof(live_range_t), live_range_cmp);
+
+   assert(mu->vregs == NULL);
+   mu->vregs = xmalloc_array(mu->num_nodes, sizeof(mir_vreg_t));
+
+   const mir_vreg_t null_vreg = { MIR_VREG_MAX, 0 };
+   for (int i = 0; i < mu->num_nodes; i++)
+      mu->vregs[i] = null_vreg;
+
+   live_range_t **active LOCAL =
+      xmalloc_array(mu->num_nodes, sizeof(live_range_t *));
+
+   SCOPED_A(mir_vreg_t) freeregs = AINIT;
+
+   // Linear scan algorithm modified to remove spilling
+   for (int i = 0, nactive = 0; i < mu->num_nodes; i++) {
+      // Expire old intervals
+      int expire = 0;
+      for (; expire < nactive; expire++) {
+         const live_range_t *next = active[expire];
+         if (next->last >= opt->ra->lrg[i].first)
+            break;
+         else {
+            assert(next->value.tag == MIR_TAG_NODE);
+            APUSH(freeregs, mu->vregs[next->value.id]);
+         }
+      }
+
+      ra_shift_active(active, 0, expire, nactive - expire);
+      nactive -= expire;
+
+      const live_range_t *this = &opt->ra->lrg[i];
+      if (mir_is_null(this->value))
+         continue;    // Does not produce value
+
+      assert(this->value.tag == MIR_TAG_NODE);
+
+      const int width = mir_get_slots(mu, mir_get_type(mu, this->value));
+
+      bool reused = false;
+      for (int j = freeregs.count - 1; j >= 0; j--) {
+         if (freeregs.items[j].width == width) {
+            mu->vregs[this->value.id] = freeregs.items[j];
+
+            if (freeregs.count > j + 1)
+               freeregs.items[j] = freeregs.items[freeregs.count - 1];
+            freeregs.count--;
+
+            reused = true;
+            break;
+         }
+      }
+
+      if (!reused) {
+         mir_vreg_t vreg = { mu->num_vregs++, width };
+         mu->vregs[this->value.id] = vreg;
+      }
+
+      // Add to active list, sorted by increasing end point
+      int pos = 0;
+      for (; pos < nactive && active[pos]->last <= opt->ra->lrg[i].last;
+           pos++)
+         assert(active[pos] != &opt->ra->lrg[i]);
+      assert(pos < mu->num_nodes);
+      if (pos < nactive)
+         ra_shift_active(active, pos + 1, pos, nactive - pos);
+      active[pos] = &opt->ra->lrg[i];
+      nactive++;
+   }
+
+   if (opt_get_verbose(OPT_RA_VERBOSE, istr(mu->name)))
+      mir_dump_optim(mu, opt);
+
+   free(opt->ra->block_start);
+   free(opt->ra);
+   opt->ra = NULL;
 }
 
 ////////////////////////////////////////////////////////////////////////////////
@@ -767,6 +1024,17 @@ static void mir_do_dce(mir_unit_t *mu, mir_optim_t *opt)
 
 // LCOV_EXCL_START /////////////////////////////////////////////////////////////
 
+static void dump_node_set(const bit_mask_t *mask, const char *tag)
+{
+   if (mask_popcount(mask) == 0)
+      return;
+
+   printf(" %s:{", tag);
+   for (size_t bit = -1, nth = 0; mask_iter(mask, &bit); nth++)
+      printf("%s%%%zd", nth++ > 0 ? "," : "", bit);
+   printf("}");
+}
+
 static int begin_block_cb(mir_unit_t *mu, mir_block_t block, int col, void *ctx)
 {
    const mir_optim_t *opt = ctx;
@@ -775,7 +1043,7 @@ static int begin_block_cb(mir_unit_t *mu, mir_block_t block, int col, void *ctx)
       return 0;
 
    const cfg_block_t *cb = &(opt->cfg[block.id]);
-   color_printf("$cyan$//");
+   nvc_printf("$cyan$//");
 
    if (cb->entry) printf(" entry");
 
@@ -801,10 +1069,15 @@ static int begin_block_cb(mir_unit_t *mu, mir_block_t block, int col, void *ctx)
          printf("%s%zd", nth > 0 ? "," : "", bit);
    }
 
+   if (cb->livein.size > 0) {
+      dump_node_set(&cb->livein, "livein");
+      dump_node_set(&cb->liveout, "liveout");
+   }
+
    if (cb->returns) printf(" returns");
    if (cb->aborts) printf(" aborts");
 
-   color_printf("$$\n%*.s", col, "");
+   nvc_printf("$$\n%*.s", col, "");
 
    return 0;
 }
@@ -812,27 +1085,51 @@ static int begin_block_cb(mir_unit_t *mu, mir_block_t block, int col, void *ctx)
 static int value_cb(mir_unit_t *mu, mir_value_t value, void *ctx)
 {
    const mir_optim_t *opt = ctx;
+   int col = 0;
 
    if (opt->gvn != NULL && gvn_tracked(value)) {
-      int col = color_printf("$!black$(");
+      col += nvc_printf("$!black$(");
 
       if (value.tag == MIR_TAG_NODE) {
          const node_data_t *n = mir_node_data(mu, value);
          col += printf("%08x:", gvn_hash_node(mu, n, opt->gvn));
       }
 
-      if (opt->gvn->nodevn[value.id] == VN_INVALID)
+      valnum_t vn = gvn_get_value(value, opt->gvn);
+      if (vn == VN_INVALID)
          col += printf("-");
       else
-         col += printf("%u", gvn_get_value(value, opt->gvn));
+         col += printf("%u", vn);
 
-      col += color_printf(")$$");
-      return col;
+      col += nvc_printf(")$$");
    }
-   else if (opt->dce != NULL && value.tag == MIR_TAG_NODE)
-      return color_printf("$!black$(%u)$$", opt->dce->uses[value.id]);
-   else
-      return 0;
+
+   if (opt->ra != NULL && value.tag == MIR_TAG_NODE) {
+      const live_range_t *lrg = NULL;
+      for (int i = 0; i < mu->num_nodes; i++) {
+         if (mir_equals(opt->ra->lrg[i].value, value)) {
+            lrg = &(opt->ra->lrg[i]);
+            break;
+         }
+      }
+
+      col += nvc_printf("$!black$[");
+
+      if (lrg == NULL)
+         col += nvc_printf("?");
+      else
+         col += nvc_printf("%u..%u", lrg->first, lrg->last);
+
+      if (mu->vregs != NULL && mu->vregs[value.id].first != MIR_VREG_MAX) {
+         col += printf(" => V%d", mu->vregs[value.id].first);
+         if (mu->vregs[value.id].width != 1)
+            col += printf("(%d)", mu->vregs[value.id].width);
+      }
+
+      col += nvc_printf("]$$");
+   }
+
+   return col;
 }
 
 static void mir_dump_optim(mir_unit_t *mu, mir_optim_t *an)
@@ -854,19 +1151,31 @@ void mir_optimise(mir_unit_t *mu, mir_pass_t passes)
 {
    mir_optim_t opt = {};
 
-   const mir_pass_t need_cfg = MIR_PASS_GVN | MIR_PASS_DCE;
+   const mir_pass_t need_dom = MIR_PASS_GVN;
+   const mir_pass_t need_liveness = MIR_PASS_DCE | MIR_PASS_RA;
+   const mir_pass_t need_cfg = need_dom | need_liveness | MIR_PASS_CFG;
 
-   if (passes & need_cfg) {
+   if (passes & need_cfg)
       opt.cfg = mir_get_cfg(mu);
+
+   if (passes & MIR_PASS_CFG)
+      mir_do_cfg_cleanup(mu, &opt);
+
+   if (passes & need_dom)
       mir_dominator_tree(mu, opt.cfg);
-   }
 
    if (passes & MIR_PASS_GVN)
       mir_do_gvn(mu, &opt);
 
+   if (passes & need_liveness)
+      mir_do_liveness(mu, &opt);
+
    if (passes & MIR_PASS_DCE)
       mir_do_dce(mu, &opt);
 
+   if (passes & MIR_PASS_RA)
+      mir_do_ra(mu, &opt);
+
    if (passes & need_cfg)
       mir_free_cfg(mu, opt.cfg);
 }
