@@ -1066,18 +1066,20 @@ static bool object_copy_mark(object_t *object, object_copy_ctx_t *ctx)
    if (arena->copygen != ctx->generation)
       return false;
 
-   if (ctx->copy_map == NULL)
-      ctx->copy_map = hash_new(1024);
+   bool marked = false;
+   if (object_marked_p(object, ctx->generation)) {
+      object_t *map = hash_get(ctx->copy_map, object);
+      if (map == object)
+         marked = true;   // Marked as root
+      else
+         return map != NULL;
+   }
 
-   if (object_marked_p(object, ctx->generation))
-      return hash_get(ctx->copy_map, object) != NULL;
+   if (!marked && ctx->should_copy[object->tag] != NULL)
+      marked = (*ctx->should_copy[object->tag])(object, ctx->pred_context);
 
    const object_class_t *class = classes[object->tag];
 
-   bool marked = false;
-   if (ctx->should_copy[object->tag] != NULL)
-      marked = (*ctx->should_copy[object->tag])(object, ctx->pred_context);
-
    object_t *copy = NULL;
    if (marked) {
       copy = object_new(global_arena, class, object->kind);
@@ -1108,6 +1110,16 @@ static bool object_copy_mark(object_t *object, object_copy_ctx_t *ctx)
    return marked;
 }
 
+void object_copy_mark_root(object_t *object, object_copy_ctx_t *ctx)
+{
+   object_arena_t *arena = __object_arena(object);
+
+   arena->copygen = ctx->generation;
+
+   if (!object_marked_p(object, ctx->generation))
+      hash_put(ctx->copy_map, object, object);
+}
+
 static object_t *object_copy_map(object_t *object, object_copy_ctx_t *ctx)
 {
    if (object == NULL)
@@ -1117,11 +1129,16 @@ static object_t *object_copy_map(object_t *object, object_copy_ctx_t *ctx)
    return map ?: object;
 }
 
-void object_copy(object_copy_ctx_t *ctx)
+void object_copy_begin(object_copy_ctx_t *ctx)
 {
+   ctx->copy_map = hash_new(1024);
+
    for (int i = 0; i < ctx->nroots; i++)
       __object_arena(ctx->roots[i])->copygen = ctx->generation;
+}
 
+void object_copy_finish(object_copy_ctx_t *ctx)
+{
    for (int i = 0; i < ctx->nroots; i++)
       (void)object_copy_mark(ctx->roots[i], ctx);
 
@@ -1132,6 +1149,7 @@ void object_copy(object_copy_ctx_t *ctx)
         hash_iter(ctx->copy_map, &it, &key, &value); ) {
       const object_t *object = key;
       object_t *copy = value;
+      assert(copy != object);
       ncopied++;
 
       copy->loc = object->loc;
@@ -1363,15 +1381,28 @@ void make_new_arena(void)
 
 void discard_global_arena(void)
 {
+   if (global_arena == NULL)
+      return;
+
    object_one_time_init();
 
-   if (global_arena != NULL) {
-      nvc_munmap(global_arena->base, global_arena->limit - global_arena->base);
+   for (void *p = global_arena->base; p != global_arena->alloc; ) {
+      assert(p < global_arena->alloc);
+      object_t *object = p;
 
-      assert(all_arenas.items[all_arenas.count - 1] == global_arena);
-      APOP(all_arenas);
+      gc_free_external(object);
 
-      free(global_arena);
-      global_arena = NULL;
+      const object_class_t *class = classes[object->tag];
+      const size_t size =
+         ALIGN_UP(class->object_size[object->kind], OBJECT_ALIGN);
+      p = (char *)p + size;
    }
+
+   nvc_munmap(global_arena->base, global_arena->limit - global_arena->base);
+
+   assert(all_arenas.items[all_arenas.count - 1] == global_arena);
+   APOP(all_arenas);
+
+   free(global_arena);
+   global_arena = NULL;
 }
