@@ -75,7 +75,12 @@ typedef enum {
    OPT_PRESERVE_CASE,
    OPT_GVN_VERBOSE,
    OPT_DCE_VERBOSE,
+   OPT_CFG_VERBOSE,
    OPT_RANDOM_SEED,
+   OPT_ELAB_STATS,
+   OPT_RELATIVE_PATH,
+   OPT_RA_VERBOSE,
+   OPT_EXCL_VERBOSE,
 
    OPT_LAST_NAME
 } opt_name_t;
