@@ -58,18 +58,33 @@ static vcode_reg_t psl_lower_boolean(lower_unit_t *lu, psl_node_t p)
       return test_reg;
 }
 
-vcode_reg_t psl_lower_union(lower_unit_t *lu, psl_node_t p)
+static vcode_reg_t lower_get_rand_bool(void)
 {
-   vcode_reg_t lhs = lower_rvalue(lu, psl_tree(psl_operand(p, 0)));
-   vcode_reg_t rhs = lower_rvalue(lu, psl_tree(psl_operand(p, 1)));
-
    ident_t func = ident_new("NVC.RANDOM.GET_RANDOM()B");
 
    vcode_reg_t context_reg = emit_link_package(ident_new("NVC.RANDOM"));
    vcode_reg_t args[] = { context_reg };
 
    vcode_type_t vbool = vtype_bool();
-   vcode_reg_t sel = emit_fcall(func, vbool, vbool, args, 1);
+   return emit_fcall(func, vbool, VCODE_INVALID_STAMP, args, 1);
+}
+
+static vcode_reg_t lower_get_rand_int(void)
+{
+   ident_t func = ident_new("NVC.RANDOM.GET_RANDOM()19NVC.RANDOM.T_UINT32");
+
+   vcode_reg_t context_reg = emit_link_package(ident_new("NVC.RANDOM"));
+   vcode_reg_t args[] = { context_reg };
+
+   vcode_type_t vint = vtype_int(0, UINT32_MAX);
+   return emit_fcall(func, vint, VCODE_INVALID_STAMP, args, 1);
+}
+
+vcode_reg_t psl_lower_union(lower_unit_t *lu, psl_node_t p)
+{
+   vcode_reg_t lhs = lower_rvalue(lu, psl_tree(psl_operand(p, 0)));
+   vcode_reg_t rhs = lower_rvalue(lu, psl_tree(psl_operand(p, 1)));
+   vcode_reg_t sel = lower_get_rand_bool();
 
    return emit_select(sel, lhs, rhs);
 }
@@ -111,8 +126,7 @@ static vcode_reg_t psl_lower_guard(lower_unit_t *lu, psl_guard_t g)
    }
 }
 
-static void psl_lower_cover(lower_unit_t *lu, psl_node_t p, cover_data_t *cover,
-                            cover_scope_t *cscope)
+static void psl_lower_cover(lower_unit_t *lu, psl_node_t p)
 {
    if (psl_has_message(p)) {
       tree_t m = psl_message(p);
@@ -127,15 +141,44 @@ static void psl_lower_cover(lower_unit_t *lu, psl_node_t p, cover_data_t *cover,
       emit_report(msg_reg, count_reg, severity_reg, locus);
    }
 
-   if (!cover_enabled(cover, COVER_MASK_FUNCTIONAL))
+   if (!cover_enabled(lu->cover, COVER_MASK_FUNCTIONAL))
       return;
 
-   cover_item_t *item = cover_add_items_for(cover, cscope, psl_to_object(p),
+   cover_item_t *item = cover_add_items_for(lu->cover, lu->cscope,
+                                            psl_to_object(p),
                                             COV_ITEM_FUNCTIONAL);
    if (item == NULL)
       return;
 
-   emit_cover_stmt(item->tag);
+   // TODO: move this to initialisation
+   vcode_reg_t counters = emit_get_counters(lu->name);
+   emit_cover_stmt(counters, item->tag);
+}
+
+static void psl_lower_assert(lower_unit_t *lu, vcode_reg_t taken_reg,
+                             vcode_reg_t locus, psl_node_t p)
+{
+   vcode_reg_t severity_reg = psl_assert_severity();
+
+   vcode_reg_t msg_reg = VCODE_INVALID_REG;
+   vcode_reg_t len_reg = VCODE_INVALID_REG;
+
+   if (psl_has_message(p)) {
+      tree_t t_msg = psl_message(p);
+      vcode_reg_t message_wrapped = lower_rvalue(lu, t_msg);
+      if (vcode_reg_kind(message_wrapped) == VCODE_TYPE_UARRAY) {
+         msg_reg = emit_unwrap(message_wrapped);
+         len_reg = emit_uarray_len(message_wrapped, 0);
+      }
+      else {
+         msg_reg = message_wrapped;
+         len_reg = emit_const(vtype_offset(), type_width(tree_type(t_msg)));
+      }
+   }
+
+   emit_assert(taken_reg, msg_reg, len_reg,
+               severity_reg, locus, VCODE_INVALID_REG,
+               VCODE_INVALID_REG);
 }
 
 static void psl_enter_state(fsm_state_t *state)
@@ -149,8 +192,7 @@ static void psl_enter_state(fsm_state_t *state)
 }
 
 static void psl_lower_state(lower_unit_t *lu, psl_fsm_t *fsm,
-                            fsm_state_t *state, vcode_block_t *state_bb,
-                            cover_data_t *cover, cover_scope_t *cscope)
+                            fsm_state_t *state, vcode_block_t *state_bb)
 {
    emit_comment("Property state %d", state->id);
 
@@ -174,14 +216,10 @@ static void psl_lower_state(lower_unit_t *lu, psl_fsm_t *fsm,
       }
 
       if (fsm->kind == FSM_COVER)
-         psl_lower_cover(lu, fsm->src, cover, cscope);
+         psl_lower_cover(lu, fsm->src);
       else if (fsm->kind == FSM_NEVER) {
-         vcode_reg_t severity_reg = psl_assert_severity();
          vcode_reg_t locus = psl_debug_locus(fsm->src);
-
-         emit_assert(vfalse, VCODE_INVALID_REG, VCODE_INVALID_REG,
-                     severity_reg, locus, VCODE_INVALID_REG,
-                     VCODE_INVALID_REG);
+         psl_lower_assert(lu, vfalse, locus, fsm->src);
       }
 
       emit_return(VCODE_INVALID_REG);
@@ -219,12 +257,8 @@ static void psl_lower_state(lower_unit_t *lu, psl_fsm_t *fsm,
    }
 
    if (fsm->kind != FSM_COVER && fsm->kind != FSM_NEVER) {
-      vcode_reg_t severity_reg = psl_assert_severity();
       vcode_reg_t locus = psl_debug_locus(state->where);
-
-      emit_assert(taken_reg, VCODE_INVALID_REG, VCODE_INVALID_REG,
-                  severity_reg, locus, VCODE_INVALID_REG,
-                  VCODE_INVALID_REG);
+      psl_lower_assert(lu, taken_reg, locus, fsm->src);
    }
 
    emit_return(VCODE_INVALID_REG);
@@ -259,7 +293,7 @@ static vcode_reg_t psl_lower_async_abort(unit_registry_t *ur,
    vcode_state_t state;
    vcode_state_save(&state);
 
-   vcode_unit_t context = get_vcode(parent);
+   vcode_unit_t context = parent->vunit;
    vcode_select_unit(context);
 
    ident_t prefix = vcode_unit_name(context);
@@ -269,7 +303,7 @@ static vcode_reg_t psl_lower_async_abort(unit_registry_t *ur,
    vcode_set_result(vtype_bool());
 
    vcode_type_t vcontext = vtype_context(prefix);
-   emit_param(vcontext, vcontext, ident_new("context"));
+   emit_param(vcontext, VCODE_INVALID_STAMP, ident_new("context"));
 
    lower_unit_t *lu = lower_unit_new(ur, parent, vu, NULL, NULL);
    unit_registry_put(ur, lu);
@@ -348,8 +382,6 @@ vcode_reg_t psl_lower_fcall(lower_unit_t *lu, psl_node_t p)
 {
    assert(psl_kind(p) == P_BUILTIN_FCALL);
 
-   tree_t expr = psl_tree(psl_operand(p, 0));
-
    switch (psl_subkind(p)) {
    case PSL_BUILTIN_PREV:
       {
@@ -361,11 +393,13 @@ vcode_reg_t psl_lower_fcall(lower_unit_t *lu, psl_node_t p)
             fatal_at(psl_loc(p), "sorry, no more than 512 cycles "
                      "are supported");
 
+         tree_t expr = psl_tree(psl_operand(p, 0));
          return psl_lower_prev_shift_reg(lu, expr, num);
       }
 
    case PSL_BUILTIN_ROSE:
       {
+         tree_t expr = psl_tree(psl_operand(p, 0));
          vcode_reg_t prev = psl_lower_prev_shift_reg(lu, expr, 1);
          vcode_reg_t prev_n = emit_not(prev);
          vcode_reg_t rhs = lower_rvalue(lu, expr);
@@ -374,6 +408,7 @@ vcode_reg_t psl_lower_fcall(lower_unit_t *lu, psl_node_t p)
 
    case PSL_BUILTIN_FELL:
       {
+         tree_t expr = psl_tree(psl_operand(p, 0));
          vcode_reg_t prev = psl_lower_prev_shift_reg(lu, expr, 1);
          vcode_reg_t expr_n = emit_not(lower_rvalue(lu, expr));
          return emit_and(prev, expr_n);
@@ -381,42 +416,134 @@ vcode_reg_t psl_lower_fcall(lower_unit_t *lu, psl_node_t p)
 
    case PSL_BUILTIN_STABLE:
       {
+         tree_t expr = psl_tree(psl_operand(p, 0));
          vcode_reg_t prev = psl_lower_prev_shift_reg(lu, expr, 1);
          vcode_reg_t rhs = lower_rvalue(lu, expr);
          return emit_cmp(VCODE_CMP_EQ, prev, rhs);
       }
 
+   case PSL_BUILTIN_NONDET:
+      {
+         assert(psl_operands(p) == 1);
+         psl_node_t vs = psl_operand(p, 0);
+
+         if (psl_subkind(vs) == PSL_VALUE_SET_BOOLEAN)
+            return lower_get_rand_bool();
+
+         int64_t candidates = 0;
+         int n_ops = psl_operands(vs);
+
+         for (int i = 0; i < n_ops; i++) {
+            psl_node_t op = psl_operand(vs, i);
+            if (psl_kind(op) == P_RANGE) {
+               const int64_t lhs = assume_int(psl_tree(psl_left(op)));
+               const int64_t rhs = assume_int(psl_tree(psl_right(op)));
+               candidates += rhs - lhs + 1;
+            }
+            else
+               candidates++;
+         }
+
+         vcode_type_t vint = vtype_int(INT64_MIN, INT64_MAX);
+
+         vcode_reg_t rnd_raw = lower_get_rand_int();
+         vcode_reg_t rnd_64bit = emit_cast(vint, VCODE_INVALID_STAMP, rnd_raw);
+         vcode_reg_t mod = emit_const(vint, candidates);
+         vcode_reg_t rnd = emit_mod(rnd_64bit, mod);
+
+         vcode_block_t test_bb[n_ops];
+         vcode_block_t ret_bb[n_ops];
+         vcode_block_t exit_bb;
+
+         for (int i = 0; i < n_ops; i++) {
+            test_bb[i] = emit_block();
+            ret_bb[i] = emit_block();
+         }
+
+         exit_bb = emit_block();
+         emit_jump(test_bb[0]);
+
+         vcode_var_t rvar = emit_var(vint, VCODE_INVALID_STAMP,
+                                     ident_new("RV"), VAR_TEMP);
+         int accum = 0;
+
+         for (int i = 0; i < n_ops; i++) {
+            psl_node_t op = psl_operand(vs, i);
+
+            vcode_select_block(test_bb[i]);
+            vcode_block_t next_bb =
+               (i < (n_ops - 1)) ? test_bb[i + 1] : exit_bb;
+
+            vcode_reg_t vaccum = emit_const(vint, accum);
+
+            if (psl_kind(op) == P_RANGE) {
+               int64_t lhs, rhs;
+               folded_int(psl_tree(psl_left(op)), &lhs);
+               folded_int(psl_tree(psl_right(op)), &rhs);
+               int64_t r = rhs - lhs;
+
+               vcode_reg_t ge = emit_cmp(VCODE_CMP_GEQ, rnd, vaccum);
+               vcode_reg_t le =
+                  emit_cmp(VCODE_CMP_LEQ, rnd, emit_const(vint, accum + r));
+               vcode_reg_t sel = emit_and(ge, le);
+
+               emit_cond(sel, ret_bb[i], next_bb);
+
+               vcode_select_block(ret_bb[i]);
+               vcode_reg_t low = emit_sub(rnd, vaccum);
+               vcode_reg_t rv = emit_add(low, emit_const(vint, lhs));
+               emit_store(rv, rvar);
+               emit_jump(exit_bb);
+
+               accum += rhs - lhs;
+            }
+            else {
+               vcode_reg_t sel = emit_cmp(VCODE_CMP_EQ, vaccum, rnd);
+               emit_cond(sel, ret_bb[i], next_bb);
+
+               vcode_select_block(ret_bb[i]);
+
+               emit_store(emit_const(vint, assume_int(psl_tree(op))), rvar);
+               emit_jump(exit_bb);
+
+               accum++;
+            }
+         }
+
+         // TODO: Handle arbitrary type conversion (e.g. all arguments of enum type)
+         vcode_select_block(exit_bb);
+         return emit_load(rvar);
+      }
+
    default:
       fatal_at(psl_loc(p), "sorry, this built-in function is not supported");
    }
 }
 
-void psl_lower_directive(unit_registry_t *ur, lower_unit_t *parent,
-                         cover_data_t *cover, tree_t wrapper)
+void psl_lower_directive(lower_unit_t *lu, object_t *obj)
 {
+   tree_t wrapper = tree_from_object(obj);
+   assert(tree_kind(wrapper) == T_PSL_DIRECT);
+
    psl_node_t p = tree_psl(wrapper);
    ident_t label = tree_ident(wrapper);
 
    psl_fsm_t *fsm = psl_fsm_new(p, label);
 
-   vcode_unit_t context = get_vcode(parent);
+   vcode_unit_t context = vcode_unit_context(lu->vunit);
 
    ident_t prefix = vcode_unit_name(context);
-   ident_t name = ident_prefix(prefix, label, '.');
+   ident_t name = vcode_unit_name(lu->vunit);
 
-   cover_scope_t *parent_cscope = lower_get_cover_scope(parent);
-   cover_scope_t *cscope = cover_create_block(cover, name, parent_cscope,
-                                              wrapper, wrapper, NULL);
-
-   vcode_unit_t vu = emit_property(name, psl_to_object(p), context);
-   lower_unit_t *lu = lower_unit_new(ur, parent, vu, NULL, NULL);
-   unit_registry_put(ur, lu);
+   lu->cscope = cover_create_block(lu->cover, name, lu->parent->cscope,
+                                   wrapper, wrapper);
 
    vcode_type_t vcontext = vtype_context(prefix);
-   emit_param(vcontext, vcontext, ident_new("context"));
+   emit_param(vcontext, VCODE_INVALID_STAMP, ident_new("context"));
 
    vcode_type_t vint32 = vtype_int(INT32_MIN, INT32_MAX);
-   vcode_reg_t state_reg = emit_param(vint32, vint32, ident_new("state"));
+   vcode_stamp_t vstamp = vstamp_int(0, INT32_MAX);
+   vcode_reg_t state_reg = emit_param(vint32, vstamp, ident_new("state"));
 
    vcode_block_t case_bb = emit_block();
    vcode_block_t abort_bb = emit_block();
@@ -451,7 +578,7 @@ void psl_lower_directive(unit_registry_t *ur, lower_unit_t *parent,
       build_wait(psl_tree(async_abort), psl_wait_cb, lu);
 
       vcode_reg_t abort_reg =
-         psl_lower_async_abort(ur, parent, async_abort, name);
+         psl_lower_async_abort(lu->registry, lu->parent, async_abort, name);
       trigger_reg = emit_or_trigger(trigger_reg, abort_reg);
    }
 
@@ -476,19 +603,18 @@ void psl_lower_directive(unit_registry_t *ur, lower_unit_t *parent,
    bool strong = false;
    for (fsm_state_t *s = fsm->states; s; s = s->next) {
       vcode_select_block(state_bb[s->id]);
-      psl_lower_state(lu, fsm, s, state_bb, cover, cscope);
+      psl_lower_state(lu, fsm, s, state_bb);
       strong |= s->strong;
    }
 
    vcode_select_block(abort_bb);
 
    if (strong) {
-      vcode_reg_t severity_reg = psl_assert_severity();
       vcode_reg_t false_reg = emit_const(vtype_bool(), 0);
       vcode_reg_t locus = psl_debug_locus(fsm->src);
-      emit_assert(false_reg, VCODE_INVALID_REG, VCODE_INVALID_REG,
-                  severity_reg, locus, VCODE_INVALID_REG,
-                  VCODE_INVALID_REG);
+
+      psl_lower_assert(lu, false_reg, locus, fsm->src);
+
       emit_return(VCODE_INVALID_REG);
    }
    else
@@ -507,8 +633,6 @@ void psl_lower_directive(unit_registry_t *ur, lower_unit_t *parent,
 
    emit_case(state_reg, abort_bb, state_ids, state_bb, fsm->next_id + 1);
 
-   unit_registry_finalise(ur, lu);
-
    psl_fsm_free(fsm);
 }
 
@@ -517,11 +641,11 @@ static void psl_lower_clock_func(lower_unit_t *lu, object_t *obj)
    psl_node_t p = psl_from_object(obj);
    assert(psl_kind(p) == P_CLOCK_DECL);
 
-   vcode_unit_t context = vcode_unit_context(get_vcode(lu));
+   vcode_unit_t context = vcode_unit_context(lu->vunit);
    ident_t prefix = vcode_unit_name(context);
 
    vcode_type_t vcontext = vtype_context(prefix);
-   emit_param(vcontext, vcontext, ident_new("context"));
+   emit_param(vcontext, VCODE_INVALID_STAMP, ident_new("context"));
 
    vcode_set_result(vtype_bool());
 
@@ -532,7 +656,7 @@ static void psl_lower_clock_func(lower_unit_t *lu, object_t *obj)
 static void psl_lower_clock_decl(unit_registry_t *ur, lower_unit_t *parent,
                                  psl_node_t p, ident_t label)
 {
-   vcode_unit_t context = get_vcode(parent);
+   vcode_unit_t context = parent->vunit;
 
    ident_t prefix = vcode_unit_name(context);
    ident_t name = ident_prefix(prefix, label, '.');
@@ -541,7 +665,7 @@ static void psl_lower_clock_decl(unit_registry_t *ur, lower_unit_t *parent,
                        NULL, psl_to_object(p));
 
    vcode_type_t vtrigger = vtype_trigger();
-   vcode_var_t var = emit_var(vtrigger, vtrigger, label, 0);
+   vcode_var_t var = emit_var(vtrigger, VCODE_INVALID_STAMP, label, 0);
 
    vcode_reg_t context_reg = emit_context_upref(0);
    vcode_reg_t args[] = { context_reg };
