@@ -1,5 +1,5 @@
 //
-//  Copyright (C) 2011-2024  Nick Gasson
+//  Copyright (C) 2011-2025  Nick Gasson
 //
 //  This program is free software: you can redistribute it and/or modify
 //  it under the terms of the GNU General Public License as published by
@@ -212,16 +212,6 @@ char *xvasprintf(const char *fmt, va_list ap) RETURNS_NONNULL;
 char *xasprintf(const char *fmt, ...)
    __attribute__((format(printf, 1, 2))) RETURNS_NONNULL;
 
-int color_printf(const char *fmt, ...)
-   __attribute__((format(printf, 1, 2)));
-int color_fprintf(FILE *file, const char *fmt, ...)
-   __attribute__((format(printf, 2, 3)));
-int color_vprintf(const char *fmt, va_list ap);
-char *color_vasprintf(const char *fmt, va_list ap);
-char *color_asprintf(const char *fmt, ...)
-   __attribute__((format(printf, 1, 2)));
-char *strip_color(const char *fmt, va_list ap);
-
 void print_centred(const char *text);
 
 void errorf(const char *fmt, ...)
@@ -281,6 +271,7 @@ bool isupper_iso88591(unsigned char ch);
 bool isalpha_iso88591(unsigned char ch);
 bool isalnum_iso88591(unsigned char ch);
 
+int checked_vsprintf(char *buf, int len, const char *fmt, va_list ap);
 int checked_sprintf(char *buf, int len, const char *fmt, ...)
    __attribute__((format(printf, 3, 4)));
 
@@ -329,6 +320,7 @@ void tb_downcase(text_buf_t *tb);
 void tb_upcase(text_buf_t *tb);
 void tb_replace(text_buf_t *tb, char old, char rep);
 void tb_strftime(text_buf_t *tb, const char *fmt, time_t time);
+void tb_ostream_write(const char *buf, size_t len, void *ctx);
 
 #define LOCAL __attribute__((cleanup(_local_free)))
 
@@ -394,6 +386,7 @@ void get_data_dir(text_buf_t *tb);
 bool get_exe_path(text_buf_t *tb);
 void open_pipe(int *rfd, int *wfd);
 bool is_absolute_path(const char *path);
+void get_relative_path(text_buf_t *tb, const char *from, const char *to);
 void check_cpu_features(void);
 
 void make_dir(const char *fmt, ...)
