@@ -155,6 +155,8 @@ unsigned vlog_size(vlog_node_t v)
          else
             return left - right + 1;
       }
+   case V_REF:
+      return vlog_size(vlog_ref(v));
    default:
       CANNOT_HANDLE(v);
    }
@@ -166,6 +168,8 @@ bool is_top_level(vlog_node_t v)
    case V_MODULE:
    case V_PRIMITIVE:
    case V_INST_BODY:
+   case V_PACKAGE:
+   case V_PROGRAM:
       return true;
    default:
       return false;
@@ -180,18 +184,30 @@ bool is_data_type(vlog_node_t v)
    case V_STRUCT_DECL:
    case V_UNION_DECL:
    case V_TYPE_DECL:
+   case V_CLASS_DECL:
       return true;
    default:
       return false;
    }
 }
 
+bool is_implicit_data_type(vlog_node_t v)
+{
+   return vlog_kind(v) == V_DATA_TYPE && vlog_subkind(v) == DT_IMPLICIT;
+}
+
 vlog_node_t vlog_longest_static_prefix(vlog_node_t v)
 {
    switch (vlog_kind(v)) {
    case V_REF:
    case V_HIER_REF:
-      return v;
+      switch (vlog_kind(vlog_ref(v))) {
+      case V_PARAM_DECL:
+      case V_LOCALPARAM:
+         return NULL;
+      default:
+         return v;
+      }
    case V_BIT_SELECT:
       {
          vlog_node_t value = vlog_value(v);
@@ -221,6 +237,8 @@ vlog_node_t vlog_longest_static_prefix(vlog_node_t v)
 
          return v;
       }
+   case V_NUMBER:
+      return NULL;
    default:
       fatal_at(vlog_loc(v), "cannot calculate longest static prefix");
    }
@@ -292,9 +310,10 @@ vlog_node_t vlog_get_type(vlog_node_t v)
    case V_VAR_DECL:
    case V_NET_DECL:
    case V_TYPE_DECL:
+   case V_CLASS_NEW:
       return vlog_type(v);
    case V_REF:
-   case V_STRUCT_REF:
+   case V_MEMBER_REF:
       if (vlog_has_ref(v))
          return vlog_get_type(vlog_ref(v));
       else
@@ -302,6 +321,41 @@ vlog_node_t vlog_get_type(vlog_node_t v)
    case V_HIER_REF:
    case V_MOD_INST:
       return NULL;
+   case V_NULL:
+      if (vlog_has_type(v))
+         return vlog_type(v);
+      else
+         return NULL;
+   default:
+      CANNOT_HANDLE(v);
+   }
+}
+
+vlog_node_t vlog_get_dim(vlog_node_t v, int n)
+{
+   switch (vlog_kind(v)) {
+   case V_REF:
+      {
+         vlog_node_t decl = vlog_ref(v), dt = vlog_type(decl);
+
+         switch (vlog_kind(decl)) {
+         case V_NET_DECL:
+         case V_VAR_DECL:
+            {
+               const int nunpacked = vlog_ranges(decl);
+               assert(n <= vlog_ranges(dt) + nunpacked);
+
+               if (n < nunpacked)
+                  return vlog_range(decl, n);
+               else
+                  return vlog_range(dt, n - nunpacked);
+            }
+         default:
+            return vlog_range(dt, n);
+         }
+      }
+   case V_BIT_SELECT:
+      return vlog_get_dim(vlog_value(v), vlog_params(v) + n);
    default:
       CANNOT_HANDLE(v);
    }
