@@ -19,20 +19,12 @@
 #define _VPI_PRIV_H
 
 #include "prim.h"
-
-#ifdef __MINGW32__
-#define PLI_DLLISPEC __declspec(dllexport)
-#define PLI_DLLESPEC
-#else
-#define PLI_DLLISPEC
-#define PLI_DLLESPEC
-#endif
-
 #include "vpi/vpi_user.h"
 
 const char *vpi_type_str(PLI_INT32 type);
 const char *vpi_method_str(PLI_INT32 type);
 const char *vpi_property_str(PLI_INT32 property);
+const char *vpi_op_type_str(PLI_INT32 subtype);
 
 __attribute__((format(printf, 3, 4)))
 void vpi_error(PLI_INT32 sev, const loc_t *loc, const char *fmt, ...);
@@ -43,6 +35,6 @@ void vpi_trace(const char *func, const char *fmt, ...);
 void vpi_register_builtins(void);
 void vpi_clear_error(void);
 void vpi_format_number(int size, const uint64_t *abits, const uint64_t *bbits,
-                       PLI_INT32 format, text_buf_t *tb);
+                       s_vpi_value *val, text_buf_t *tb);
 
 #endif  // _VPI_PRIV_H
