@@ -21,6 +21,7 @@
 #include "ident.h"
 #include "option.h"
 #include "phase.h"
+#include "printf.h"
 #include "psl/psl-node.h"
 #include "psl/psl-phase.h"
 #include "scan.h"
@@ -43,8 +44,8 @@ static void diff_dump(const char *have, const char *expect)
       const int len2 = nl2 - expect;
 
       if (len1 != len2 || strncmp(have, expect, len1) != 0) {
-         color_printf("$red$- %.*s$$\n", len2, expect);
-         color_printf("$green$+ %.*s$$\n", len1, have);
+         nvc_printf("$red$- %.*s$$\n", len2, expect);
+         nvc_printf("$green$+ %.*s$$\n", len1, have);
          ck_abort_msg("found difference in dump");
       }
       else if (*nl1 == '\0' || *nl2 == '\0')
@@ -184,6 +185,16 @@ START_TEST(test_vhdl2)
         "begin\n"
         "  assert \">\"(R, 0.000000);\n"
         "end generate;\n" },
+      { "P2",
+        "P2: process (S2) is\n"
+        "begin\n"
+        "  case S2 is\n"
+        "    when 0 =>\n"
+        "      null;\n"
+        "    when 2 to 4 | others =>\n"
+        "      null;\n"
+        "  end case;\n"
+        "end process;\n" },
    };
 
    const int nstmts = tree_stmts(a);
@@ -491,8 +502,7 @@ START_TEST(test_vlog1)
              "  input d;\n"
              "  input clk;\n"
              "  input rstb;\n"
-             "  output q;\n"
-             "  logic q;\n"
+             "  output logic q;\n"
              "  always @(posedge clk) q <= d;\n"
              "endmodule // dff\n\n");
    tb_rewind(tb);
@@ -503,17 +513,18 @@ START_TEST(test_vlog1)
    vlog_dump(m2, 0);
    diff_dump(tb_get(tb),
              "module mod2;\n"
-             "  wire logic [7:0] bus;\n"
-             "  wire logic w;\n"
+             "  wire /* implicit */ [7:0] bus;\n"
+             "  wire /* implicit */ w;\n"
              "  logic r;\n"
              "  integer i;\n"
-             "  parameter logic [7:0] p1 = 5;\n"
+             "  parameter /* implicit */ [7:0] p1 = 5;\n"
              "  localparam bit p2 = 0;\n"
              "  logic [7:0] array1 [127:0];\n"
              "  function func;\n"
              "    input p;\n"
              "  endfunction\n"
              "  enum byte { a=5, b } e1;\n"
+             "  logic signed [3:0] s1;\n"
              "  initial begin\n"
              "    $display(\"hello\", $time);\n"
              "    if (bus) r <= 1 | r;\n"
@@ -569,9 +580,9 @@ START_TEST(test_vlog1)
    vlog_dump(m4, 0);
    diff_dump(tb_get(tb),
              "module mod3;\n"
-             "  wire logic x;\n"
-             "  wire logic y;\n"
-             "  wire logic z;\n"
+             "  wire /* implicit */ x;\n"
+             "  wire /* implicit */ y;\n"
+             "  wire /* implicit */ z;\n"
              "  real v1 = 1.5;\n"
              "  integer v2 = x ? 1 : 2;\n"
              "  wire logic zz;\n"
