@@ -21,57 +21,14 @@
 #include "mir/mir-unit.h"
 #include "object.h"
 #include "option.h"
-
-#include <inttypes.h>
-
-typedef union {
-   struct {
-      uint64_t data : 60;
-      uint64_t tag : 4;
-   };
-   struct {
-      float    real;
-      uint32_t ones : 28;
-      uint32_t zero : 4;
-   };
-   const void *ptr;
-   uint64_t    bits;
-} mir_pattern_t;
-
-STATIC_ASSERT(sizeof(mir_pattern_t) == 8);
-
-#define _ (UINT32_MAX + 1)
-#define VAR(name) ((mir_pattern_t){ .ptr = ("\x8" name) })
-#define PARAM(name) ((mir_pattern_t){ .ptr = ("\x2" name) })
-#define LINK(name) ((mir_pattern_t){ .ptr = ("\xa" name) })
-#define NODE(n) ((mir_pattern_t){ .tag = MIR_TAG_NODE, .data = n })
-#define BLOCK(n) ((mir_pattern_t){ .tag = MIR_TAG_BLOCK, .data = n })
-#define CONST(n) ((mir_pattern_t){ .tag = MIR_TAG_CONST, .data = n })
-#define ENUM(n) ((mir_pattern_t){ .tag = MIR_TAG_ENUM, .data = n })
-#define REAL(f) ((mir_pattern_t){ .ones = ~0, .real = (float)(f) })
-
-#define mir_match(mu, b, pat) _mir_match((mu), (b), (pat), ARRAY_LEN((pat)))
-
-#define mir_assert_const_eq(mu, value, exp) do {          \
-      int64_t __cval;                                     \
-      ck_assert(mir_get_const((mu), (value), &__cval));   \
-      ck_assert_int_eq(__cval, (exp));                    \
-   } while (0)
-
-typedef struct {
-   mir_op_t      op;
-   mir_pattern_t arg0;
-   mir_pattern_t arg1;
-   mir_pattern_t arg2;
-} mir_match_t;
-
-STATIC_ASSERT(sizeof(mir_match_t) <= 32);
+#include "test_mir.h"
 
 static const char *mir_tag_name(int tag)
 {
    static const char *map[] = {
       [MIR_TAG_PARAM] = "parameter",
       [MIR_TAG_VAR] = "variable",
+      [MIR_TAG_EXTVAR] = "external variable",
       [MIR_TAG_LINKAGE] = "linkage",
       [MIR_TAG_BLOCK] = "block",
       [MIR_TAG_NODE] = "node",
@@ -114,7 +71,7 @@ static void mir_match_arg(mir_unit_t *mu, mir_value_t node,
       }
    }
    else if (tag == MIR_TAG_VAR || tag == MIR_TAG_PARAM
-            || tag == MIR_TAG_LINKAGE) {
+            || tag == MIR_TAG_LINKAGE || tag == MIR_TAG_EXTVAR) {
       ck_assert_ptr_nonnull(str);
 
       if (actual.tag != tag) {
@@ -160,8 +117,7 @@ static void mir_match_arg(mir_unit_t *mu, mir_value_t node,
    }
 }
 
-static void _mir_match(mir_unit_t *mu, int nth, const mir_match_t *mm,
-                       size_t length)
+void _mir_match(mir_unit_t *mu, int nth, const mir_match_t *mm, size_t length)
 {
    ck_assert_int_lt(nth, mir_count_blocks(mu));
    mir_block_t block = mir_get_block(mu, nth);
@@ -1164,6 +1120,39 @@ START_TEST(test_check1)
 }
 END_TEST
 
+START_TEST(test_cfg1)
+{
+   mir_unit_t *mu = mir_unit_new(get_mir(), ident_new("cfg1"), NULL,
+                                 MIR_UNIT_FUNCTION, NULL);
+
+   mir_build_return(mu, MIR_NULL_VALUE);
+
+   mir_block_t b2 = mir_add_block(mu);
+   mir_block_t b3 = mir_add_block(mu);
+
+   mir_set_cursor(mu, b2, MIR_APPEND);
+
+   mir_type_t t_offset = mir_offset_type(mu);
+
+   mir_build_debug_out(mu, mir_const(mu, t_offset, 42));
+   mir_build_jump(mu, b3);
+
+   mir_set_cursor(mu, b3, MIR_APPEND);
+
+   mir_build_debug_out(mu, mir_const(mu, t_offset, 111));
+   mir_build_jump(mu, b2);
+
+   mir_optimise(mu, MIR_PASS_CFG);
+
+   for (int i = 1; i <= 2; i++) {
+      static const mir_match_t bb[] = {
+         { MIR_OP_UNREACHABLE },
+      };
+      mir_match(mu, i, bb);
+   }
+}
+END_TEST
+
 Suite *get_mir_tests(void)
 {
    Suite *s = suite_create("mir");
@@ -1191,6 +1180,7 @@ Suite *get_mir_tests(void)
    tcase_add_test(tc, test_vec1);
    tcase_add_test(tc, test_vec2);
    tcase_add_test(tc, test_check1);
+   tcase_add_test(tc, test_cfg1);
    suite_add_tcase(s, tc);
 
    return s;
